"""
LatLon holds a latitude and longitude location and provides several helper
functions.
"""

class LatLon(object):
    __slots__ = [ 'lat', 'lon' ]
    def __init__(self, lat: float, lon: float):
        self.lat : float = lat
        self.lon : float = lon
        # if not utils.latlon_in_zone(self, 'B') and not utils.latlon_in_zone(self, 'D'):
        #     raise ValueError(f'{self} outside of project boundary')

    def __eq__(self, other: any):
        if isinstance(other, LatLon):
            return self.lat == other.lat and self.lon == other.lon
        return False

    def __repr__(self):
        return f'LatLon({self.lat}, {self.lon})'

    def __str__(self):
        return f'{self.lat:.2f}:{self.lon:.2f}'


    def to_tuple(self):
        return (self.lat, self.lon)


    # TODO is this used?
    def to_dims(self):
        return {
                'lat': self.lat,
                'lon': self.lon,
                }


    def get_label(self):
        return f'({self.lat:.2f}, {self.lon:.2f})'

    @staticmethod
    def from_string(string: str):
        import re
        # [lat, lon] = string.split(',')
        [lat, lon] = re.split(r'[:,_]', string)
        lat = float(lat)
        lon = -float(lon)
        return LatLon(lat, lon)

    @staticmethod
    def Portland():
        # return LatLon(45.5155, -122.668)
        return LatLon(45.5051, -122.6750)



zones = ['A', 'B', 'C', 'D', 'E']
realization_numbers = [
        2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
        17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27,
        28, 29, 30, 31, 32, 33
        ]
realization_strings = [
        'mean',
        'std',
        'std-'
        ]
realization_keys = [
        *realization_numbers,
        *realization_strings,
        ]

class Realization(object):
    __slots__ = [ 'zone', 'key' ]
    def __init__(self,
            zone: str,
            key: [int, str]):
        if zone not in zones:
            raise ValueError('Invalid realization zone')
        if key not in realization_keys:
            raise ValueError('Invalid realization key')
        self.zone = zone
        self.key = key

    @staticmethod
    def valid_zones():
        return zones.copy()

    @staticmethod
    def valid_realization_numbers():
        return realization_numbers.copy()


    @staticmethod
    def from_string(string: str):
        """
        Creates a realization from a string like: C27 or Dmean
        """
        zone = string[0]
        key = ''
        try:
            key = int(string[1:])
        except:
            key = string[1:]
        return Realization(zone, key)

    def into_string(self):
        return self.__str__()

    def __str__(self):
        return f'{self.zone}{self.key}'

    @property
    def file_path(self):
        from os import path
        if self.key in realization_keys:
            return path.join(self.folder_name, self.zone, 'Xarray.nc')

    @property
    def folder_name(self):
        from os import path
        if self.key == 2:
            return 'csz002_sd10'
        elif self.key in realization_numbers:
            return f'csz{self.key:03d}'
        elif self.key in realization_strings:
            return path.join('All', self.key)
        raise f'Key "{self.key}" has no associated folder name'

    @property
    def has_m9(self):
        return self.key in realization_numbers



import os
import json

rootdir = os.path.dirname(os.path.abspath(__file__))

config = json.load(open(os.path.join(rootdir, 'appdata', 'config.json')))
data_format = json.load(open(os.path.join(rootdir, config['paths']['data_format'])))



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
        name='istar_m9csz_core',
        version='0.0.4',
        author="Matt O'Tousa",
        author_email='motousa@pdx.edu',
        description='Core M9CSZ data parsing',
        long_description_content_type='text/markdown',
        url='https://github.com/Tattomoosa/m9csz',
        packages=setuptools.find_packages(),
        # data_files=[
            # ('appdata', ['istar_m9csz_core/appdata/*.json'])
            # ],
        install_requires=[
            'dask', 'netCDF4', 'numpy', 'pandas',
            'six', 'xarray'
            ],
        classifiers=[
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            ],
        python_requires='>=3.7')



import xarray
from typing import List
from .structs.realization import Realization
from .structs.latlon import LatLon
from .appdata import config, data_format
import numpy as np

"""
A wrapper around an xarray object exposing easy ways to visualize a collection
of realization's data either as a map or a line.
"""

class Data(object):

    # Must be set externally, with e.g. Data.m9_path = 'a/path'
    m9_path : str = None
    m9csz_path : str = None
    z_path : str = None

    def __init__(self,
            realizations : [Realization, List[Realization]]):
        # Error if no path supplied to class
        if Data.m9_path is None\
                or Data.m9csz_path is None\
                or Data.m9csz_path is None:
            raise ValueError('Data.m9_path, Data.m9csz_path, Data.z_path must be set')
        # Always turn into list
        if not isinstance(realizations, list):
            realizations = [realizations]
        # Load Xarray data
        self.__data = self.__load_data(realizations)

    def __load_data(self, realizations: List[Realization]) -> List[str]:
        def get_paths(
                realizations : List[Realization],
                project_path : str) -> List[str] :
            """Gets path on disk for each realization"""
            from os import path
            base : str = project_path
            # return [path.join(base, r.folder_name, r.zone, 'Xarray.nc')
            return [path.join(base, r.file_path) for r in realizations]

        def get_data(
                realizations : List[Realization],
                project_path : str) -> xarray.Dataset :
            """Loads xarray data from disk"""
            realizations_xarray = xarray.DataArray(
                    [r.into_string() for r in realizations],
                    name='realization',
                    dims={'realization': realizations})
            # Build paths for each realization
            paths : List[str] = get_paths(realizations, project_path)
            # Get data
            return xarray.open_mfdataset(
                    paths,
                    coords='minimal',
                    combine='nested',
                    concat_dim=[realizations_xarray])

        m9_realizations = [r for r in realizations if r.has_m9]
        print(m9_realizations)

        # Load M9 data
        m9_data = None
        if m9_realizations:
            m9_data = get_data(m9_realizations, Data.m9_path)
        # Prepend `m9_` to m9 spectral data, so m9csz data names don't conflict
        if m9_data:
            rename_fields = ['period', 'spectra_x', 'spectra_y']
            m9_data = m9_data.rename({d : f'm9_{d}' for d in rename_fields})
        try:
            # Attempt to load M9CSZ data
            try:
                m9csz_data = get_data(realizations, Data.m9csz_path)
            except:
                raise ValueError('Error loading m9csz data')
            if m9_data:
                try:
                    return xarray.combine_by_coords(
                            [m9_data, m9csz_data],
                            join='outer')
                except: raise ValueError('Error combining m9 and m9csz data.')
            return m9csz_data

        except ValueError as e:
            # On failure, use M9 data instead of halting
            print(e)
            print('Using m9 data...')
            return m9_data

    @property
    def data(self):
        return self.__data

    def get_map(self,
            field: str,
            time: float,
            ):
        if field is None:
            raise ValueError('Must specify a map field')
        if time is None:
            time = 0.0
        data = self.__data
        try:
            time_axis = data_format['xarray']['m9csz']['fields'][field]['line_axis']
            if time_axis == 'time':
                data = data.where(data.time < data.CutOffTime)
            data = data[field].sel({ time_axis: time }, method='nearest', drop=True)
        except:
            data = data[field]
        data = np.abs(data)
        data = data.mean(dim='realization')
        data = data.squeeze()
        return data

    def get_line(self,
            field: str,
            latlon: LatLon):
        time_axis = data_format['xarray']['m9csz']['fields'][field]['line_axis']
        data = self.__data
        if time_axis == 'time':
            data = data.where(data.time < data.CutOffTime)
        data = data[field].sel(
                { 'lat': latlon.lat, 'lon': latlon.lon },
                method='nearest',
                drop=True)
        data = data.mean(dim='realization')
        data = data.squeeze()
        return data
